@@ -1,4 +1,4 @@
-# $Id: io.py 9100 2022-07-04 21:06:49Z milde $
+# $Id: io.py 10146 2025-05-27 06:14:22Z milde $
 # Author: David Goodger <goodger@python.org>
 # Copyright: This module has been placed in the public domain.
 
@@ -7,6 +7,8 @@ I/O classes provide a uniform API for low-level input and output.  Subclasses
 exist for a variety of input/output mechanisms.
 """
 
+from __future__ import annotations
+
 __docformat__ = 'reStructuredText'
 
 import codecs
@@ -18,6 +20,12 @@ import warnings
 
 from docutils import TransformSpec
 
+TYPE_CHECKING = False
+if TYPE_CHECKING:
+    from typing import Any, BinaryIO, ClassVar, Final, Literal, TextIO
+
+    from docutils import nodes
+    from docutils.nodes import StrPath
 
 # Guess the locale's preferred encoding.
 # If no valid guess can be made, _locale_encoding is set to `None`:
@@ -27,17 +35,13 @@ from docutils import TransformSpec
 #       before importing this module
 try:
     # Return locale encoding also in UTF-8 mode
-    _locale_encoding = locale.getlocale()[1] or locale.getdefaultlocale()[1]
-    _locale_encoding = _locale_encoding.lower()
-except ValueError as error:  # OS X may set UTF-8 without language code
-    # See https://bugs.python.org/issue18378 fixed in 3.8
-    # and https://sourceforge.net/p/docutils/bugs/298/.
-    # Drop the special case after requiring Python >= 3.8
-    if "unknown locale: UTF-8" in error.args:
-        _locale_encoding = "utf-8"
-    else:
-        _locale_encoding = None
-except:  # noqa  any other problems determining the locale -> use None
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        _locale_encoding: str | None = (locale.getlocale()[1]
+                                        or locale.getdefaultlocale()[1]
+                                        ).lower()
+except:  # NoQA: E722 (catchall)
+    # Any problem determining the locale: use None
     _locale_encoding = None
 try:
     codecs.lookup(_locale_encoding)
@@ -49,7 +53,7 @@ class InputError(OSError): pass
 class OutputError(OSError): pass
 
 
-def check_encoding(stream, encoding):
+def check_encoding(stream: TextIO, encoding: str) -> bool | None:
     """Test, whether the encoding of `stream` matches `encoding`.
 
     Returns
@@ -65,24 +69,35 @@ def check_encoding(stream, encoding):
         return None
 
 
-def error_string(err):
+def error_string(err: BaseException) -> str:
     """Return string representation of Exception `err`.
     """
     return f'{err.__class__.__name__}: {err}'
 
 
 class Input(TransformSpec):
-
     """
     Abstract base class for input wrappers.
+
+    Docutils input objects must provide a `read()` method that
+    returns the source, typically as `str` instance.
+
+    Inheriting `TransformSpec` allows input objects to add "transforms" to
+    the "Transformer".  (Since Docutils 0.19, input objects are no longer
+    required to be `TransformSpec` instances.)
     """
 
-    component_type = 'input'
+    component_type: Final = 'input'
 
-    default_source_path = None
+    default_source_path: ClassVar[str | None] = None
 
-    def __init__(self, source=None, source_path=None, encoding=None,
-                 error_handler='strict'):
+    def __init__(
+        self,
+        source: str | TextIO | nodes.document | None = None,
+        source_path: StrPath | None = None,
+        encoding: str | Literal['unicode'] | None = 'utf-8',
+        error_handler: str | None = 'strict',
+    ) -> None:
         self.encoding = encoding
         """Text encoding for the input source."""
 
@@ -101,31 +116,34 @@ class Input(TransformSpec):
         self.successful_encoding = None
         """The encoding that successfully decoded the source data."""
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return '%s: source=%r, source_path=%r' % (self.__class__, self.source,
                                                   self.source_path)
 
-    def read(self):
+    def read(self) -> str:
+        """Return input as `str`. Define in subclasses."""
         raise NotImplementedError
 
-    def decode(self, data):
+    def decode(self, data: str | bytes) -> str:
         """
         Decode `data` if required.
 
         Return Unicode `str` instances unchanged (nothing to decode).
 
         If `self.encoding` is None, determine encoding from data
-        or try UTF-8, locale encoding, and (as last ressort) 'latin-1'.
-        The client application should call ``locale.setlocale`` at the
+        or try UTF-8 and the locale's preferred encoding.
+        The client application should call ``locale.setlocale()`` at the
         beginning of processing::
 
             locale.setlocale(locale.LC_ALL, '')
 
         Raise UnicodeError if unsuccessful.
+
+        Provisional: encoding detection will be removed in Docutils 1.0.
         """
         if self.encoding and self.encoding.lower() == 'unicode':
             assert isinstance(data, str), ('input encoding is "unicode" '
-                                           'but input is not a `str` object')
+                                           'but `data` is no `str` instance')
         if isinstance(data, str):
             # nothing to decode
             return data
@@ -134,30 +152,33 @@ class Input(TransformSpec):
             # explicitly given.
             encoding_candidates = [self.encoding]
         else:
-            data_encoding = self.determine_encoding_from_data(data)
+            with warnings.catch_warnings():
+                warnings.filterwarnings('ignore', category=DeprecationWarning)
+                data_encoding = self.determine_encoding_from_data(data)
             if data_encoding:
-                # If the data declares its encoding (explicitly or via a BOM),
-                # we believe it.
+                # `data` declares its encoding with  "magic comment" or BOM,
                 encoding_candidates = [data_encoding]
             else:
-                # Apply heuristics only if no encoding is explicitly given and
-                # no BOM found.  Start with UTF-8, because that only matches
+                # Apply heuristics if the encoding is not specified.
+                # Start with UTF-8, because that only matches
                 # data that *IS* UTF-8:
                 encoding_candidates = ['utf-8']
-                # TODO: use `locale.getpreferredlocale(do_setlocale=True)`
-                # to respect UTF-8 mode (API change).
-                # (Check if it is a valid encoding and not UTF-8)
-                if _locale_encoding and _locale_encoding != 'utf-8':
-                    encoding_candidates.append(_locale_encoding)
-                # TODO: don't fall back to 'latin-1' (API change).
-                encoding_candidates.append('latin-1')
+                # If UTF-8 fails, fall back to the locale's preferred encoding:
+                if sys.version_info[:2] >= (3, 11):
+                    fallback = locale.getencoding()
+                else:
+                    fallback = locale.getpreferredencoding(do_setlocale=False)
+                if fallback and fallback.lower() != 'utf-8':
+                    encoding_candidates.append(fallback)
+        if not self.encoding and encoding_candidates[0] != 'utf-8':
+            warnings.warn('Input encoding auto-detection will be removed and '
+                          'the encoding values None and "" become invalid '
+                          'in Docutils 1.0.', DeprecationWarning, stacklevel=2)
         for enc in encoding_candidates:
             try:
                 decoded = str(data, enc, self.error_handler)
                 self.successful_encoding = enc
-                # Return decoded, removing BOM and other ZWNBSPs.
-                # TODO: only remove BOM (ZWNBSP at start of data, API change).
-                return decoded.replace('\ufeff', '')
+                return decoded
             except (UnicodeError, LookupError) as err:
                 # keep exception instance for use outside of the "for" loop.
                 error = err
@@ -166,21 +187,35 @@ class Input(TransformSpec):
             f'{", ".join(repr(enc) for enc in encoding_candidates)}.\n'
             f'({error_string(error)})')
 
-    coding_slug = re.compile(br"coding[:=]\s*([-\w.]+)")
+    coding_slug: ClassVar[re.Pattern[bytes]] = re.compile(
+        br'coding[:=]\s*([-\w.]+)'
+    )
     """Encoding declaration pattern."""
 
-    byte_order_marks = ((codecs.BOM_UTF8, 'utf-8'),
-                        (codecs.BOM_UTF16_BE, 'utf-16-be'),
-                        (codecs.BOM_UTF16_LE, 'utf-16-le'),)
+    byte_order_marks: ClassVar[tuple[tuple[bytes, str], ...]] = (
+        (codecs.BOM_UTF32_BE, 'utf-32'),
+        (codecs.BOM_UTF32_LE, 'utf-32'),
+        (codecs.BOM_UTF8, 'utf-8-sig'),
+        (codecs.BOM_UTF16_BE, 'utf-16'),
+        (codecs.BOM_UTF16_LE, 'utf-16'),
+    )
     """Sequence of (start_bytes, encoding) tuples for encoding detection.
     The first bytes of input data are checked against the start_bytes strings.
-    A match indicates the given encoding."""
+    A match indicates the given encoding.
 
-    def determine_encoding_from_data(self, data):
+    Internal. Will be removed in Docutils 1.0.
+    """
+
+    def determine_encoding_from_data(self, data: bytes) -> str | None:
         """
         Try to determine the encoding of `data` by looking *in* `data`.
         Check for a byte order mark (BOM) or an encoding declaration.
+
+        Deprecated. Will be removed in Docutils 1.0.
         """
+        warnings.warn('docutils.io.Input.determine_encoding_from_data() '
+                      'will be removed in Docutils 1.0.',
+                      DeprecationWarning, stacklevel=2)
         # check for a byte order mark:
         for start_bytes, encoding in self.byte_order_marks:
             if data.startswith(start_bytes):
@@ -192,7 +227,8 @@ class Input(TransformSpec):
                 return match.group(1).decode('ascii')
         return None
 
-    def isatty(self):
+    def isatty(self) -> bool:
+        """Return True, if the input source is connected to a TTY device."""
         try:
             return self.source.isatty()
         except AttributeError:
@@ -200,45 +236,64 @@ class Input(TransformSpec):
 
 
 class Output(TransformSpec):
-
     """
     Abstract base class for output wrappers.
+
+    Docutils output objects must provide a `write()` method that
+    expects and handles one argument (the output).
+
+    Inheriting `TransformSpec` allows output objects to add "transforms" to
+    the "Transformer".  (Since Docutils 0.19, output objects are no longer
+    required to be `TransformSpec` instances.)
     """
 
-    component_type = 'output'
+    component_type: Final = 'output'
 
-    default_destination_path = None
+    default_destination_path: ClassVar[str | None] = None
 
-    def __init__(self, destination=None, destination_path=None,
-                 encoding=None, error_handler='strict'):
-        self.encoding = encoding
+    def __init__(
+        self,
+        destination: TextIO | str | bytes | None = None,
+        destination_path: StrPath | None = None,
+        encoding: str | None = None,
+        error_handler: str | None = 'strict',
+    ) -> None:
+        self.encoding: str | None = encoding
         """Text encoding for the output destination."""
 
-        self.error_handler = error_handler or 'strict'
+        self.error_handler: str = error_handler or 'strict'
         """Text encoding error handler."""
 
-        self.destination = destination
+        self.destination: TextIO | str | bytes | None = destination
         """The destination for output data."""
 
-        self.destination_path = destination_path
+        self.destination_path: StrPath | None = destination_path
         """A text reference to the destination."""
 
         if not destination_path:
             self.destination_path = self.default_destination_path
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return ('%s: destination=%r, destination_path=%r'
                 % (self.__class__, self.destination, self.destination_path))
 
-    def write(self, data):
-        """`data` is a Unicode string, to be encoded by `self.encode`."""
+    def write(self, data: str | bytes) -> str | bytes | None:
+        """Write `data`. Define in subclasses."""
         raise NotImplementedError
 
-    def encode(self, data):
+    def encode(self, data: str | bytes) -> str | bytes:
+        """
+        Encode and return `data`.
+
+        If `data` is a `bytes` instance, it is returned unchanged.
+        Otherwise it is encoded with `self.encoding`.
+
+        Provisional: If `self.encoding` is set to the pseudo encoding name
+        "unicode", `data` must be a `str` instance and is returned unchanged.
+        """
         if self.encoding and self.encoding.lower() == 'unicode':
-            assert isinstance(data, str), (
-                'the encoding given is "unicode" but the output is not '
-                'a Unicode string')
+            assert isinstance(data, str), ('output encoding is "unicode" '
+                                           'but `data` is no `str` instance')
             return data
         if not isinstance(data, str):
             # Non-unicode (e.g. bytes) output.
@@ -250,13 +305,16 @@ class Output(TransformSpec):
 class ErrorOutput:
     """
     Wrapper class for file-like error streams with
-    failsafe de- and encoding of `str`, `bytes`, `unicode` and
-    `Exception` instances.
+    failsafe de- and encoding of `str`, `bytes`, and `Exception` instances.
     """
 
-    def __init__(self, destination=None, encoding=None,
-                 encoding_errors='backslashreplace',
-                 decoding_errors='replace'):
+    def __init__(
+        self,
+        destination: TextIO | BinaryIO | str | Literal[False] | None = None,
+        encoding: str | None = None,
+        encoding_errors: str = 'backslashreplace',
+        decoding_errors: str = 'replace',
+    ) -> None:
         """
         :Parameters:
             - `destination`: a file-like object,
@@ -274,20 +332,24 @@ class ErrorOutput:
         elif isinstance(destination, str):
             destination = open(destination, 'w')
 
-        self.destination = destination
+        self.destination: TextIO | BinaryIO | Literal[False] = destination
         """Where warning output is sent."""
 
-        self.encoding = (encoding or getattr(destination, 'encoding', None)
-                         or _locale_encoding or 'ascii')
+        self.encoding: str = (
+            encoding
+            or getattr(destination, 'encoding', None)
+            or _locale_encoding
+            or 'ascii'
+        )
         """The output character encoding."""
 
-        self.encoding_errors = encoding_errors
+        self.encoding_errors: str = encoding_errors
         """Encoding error handler."""
 
-        self.decoding_errors = decoding_errors
+        self.decoding_errors: str = decoding_errors
         """Decoding error handler."""
 
-    def write(self, data):
+    def write(self, data: str | bytes | Exception) -> None:
         """
         Write `data` to self.destination. Ignore, if self.destination is False.
 
@@ -297,23 +359,29 @@ class ErrorOutput:
             return
         if isinstance(data, Exception):
             data = str(data)
+        # The destination is either opened in text or binary mode.
+        # If data has the wrong type, try to convert it.
         try:
             self.destination.write(data)
         except UnicodeEncodeError:
-            self.destination.write(data.encode(self.encoding,
-                                               self.encoding_errors))
+            # Encoding data from string to bytes failed with the
+            # destination's encoding and error handler.
+            # Try again with our own encoding and error handler.
+            binary = data.encode(self.encoding, self.encoding_errors)
+            self.destination.write(binary)
         except TypeError:
             if isinstance(data, str):  # destination may expect bytes
-                self.destination.write(data.encode(self.encoding,
-                                                   self.encoding_errors))
+                binary = data.encode(self.encoding, self.encoding_errors)
+                self.destination.write(binary)
             elif self.destination in (sys.stderr, sys.stdout):
                 # write bytes to raw stream
                 self.destination.buffer.write(data)
             else:
-                self.destination.write(str(data, self.encoding,
-                                           self.decoding_errors))
+                # destination in text mode, write str
+                string = data.decode(self.encoding, self.decoding_errors)
+                self.destination.write(string)
 
-    def close(self):
+    def close(self) -> None:
         """
         Close the error-output stream.
 
@@ -327,7 +395,8 @@ class ErrorOutput:
         except AttributeError:
             pass
 
-    def isatty(self):
+    def isatty(self) -> bool:
+        """Return True, if the destination is connected to a TTY device."""
         try:
             return self.destination.isatty()
         except AttributeError:
@@ -339,14 +408,20 @@ class FileInput(Input):
     """
     Input for single, simple file-like objects.
     """
-    def __init__(self, source=None, source_path=None,
-                 encoding=None, error_handler='strict',
-                 autoclose=True, mode='r'):
+    def __init__(
+        self,
+        source: TextIO | None = None,
+        source_path: StrPath | None = None,
+        encoding: str | Literal['unicode'] | None = 'utf-8',
+        error_handler: str | None = 'strict',
+        autoclose: bool = True,
+        mode: Literal['r', 'rb', 'br'] = 'r'
+    ) -> None:
         """
         :Parameters:
             - `source`: either a file-like object (which is read directly), or
               `None` (which implies `sys.stdin` if no `source_path` given).
-            - `source_path`: a path to a file, which is opened and then read.
+            - `source_path`: a path to a file, which is opened for reading.
             - `encoding`: the expected text encoding of the input file.
             - `error_handler`: the encoding error handler to use.
             - `autoclose`: close automatically after read (except when
@@ -354,7 +429,7 @@ class FileInput(Input):
             - `mode`: how the file is to be opened (see standard function
               `open`). The default is read only ('r').
         """
-        Input.__init__(self, source, source_path, encoding, error_handler)
+        super().__init__(source, source_path, encoding, error_handler)
         self.autoclose = autoclose
         self._stderr = ErrorOutput()
 
@@ -362,7 +437,7 @@ class FileInput(Input):
             if source_path:
                 try:
                     self.source = open(source_path, mode,
-                                       encoding=self.encoding or 'utf-8-sig',
+                                       encoding=self.encoding,
                                        errors=self.error_handler)
                 except OSError as error:
                     raise InputError(error.errno, error.strerror, source_path)
@@ -379,57 +454,57 @@ class FileInput(Input):
             except AttributeError:
                 pass
 
-    def read(self):
+    def read(self) -> str:
         """
-        Read and decode a single file and return the data (Unicode string).
+        Read and decode a single file, return as `str`.
         """
         try:
-            if self.source is sys.stdin:
-                # read as binary data to circumvent auto-decoding
+            if not self.encoding and hasattr(self.source, 'buffer'):
+                # read as binary data
                 data = self.source.buffer.read()
+                # decode with heuristics
+                data = self.decode(data)
+                # normalize newlines
+                data = '\n'.join(data.splitlines()+[''])
             else:
                 data = self.source.read()
-        except (UnicodeError, LookupError):
-            if not self.encoding and self.source_path:
-                # re-read in binary mode and decode with heuristics
-                b_source = open(self.source_path, 'rb')
-                data = b_source.read()
-                b_source.close()
-            else:
-                raise
         finally:
             if self.autoclose:
                 self.close()
-        data = self.decode(data)
-        # normalise newlines
-        return '\n'.join(data.splitlines()+[''])
+        return data
 
-    def readlines(self):
+    def readlines(self) -> list[str]:
         """
-        Return lines of a single file as list of Unicode strings.
+        Return lines of a single file as list of strings.
         """
         return self.read().splitlines(True)
 
-    def close(self):
+    def close(self) -> None:
         if self.source is not sys.stdin:
             self.source.close()
 
 
 class FileOutput(Output):
 
-    """
-    Output for single, simple file-like objects.
-    """
+    """Output for single, simple file-like objects."""
 
-    mode = 'w'
+    default_destination_path: Final = '<file>'
+
+    mode: Literal['w', 'a', 'x', 'wb', 'ab', 'xb', 'bw', 'ba', 'bx'] = 'w'
     """The mode argument for `open()`."""
     # 'wb' for binary (e.g. OpenOffice) files (see also `BinaryFileOutput`).
     # (Do not use binary mode ('wb') for text files, as this prevents the
     # conversion of newlines to the system specific default.)
 
-    def __init__(self, destination=None, destination_path=None,
-                 encoding=None, error_handler='strict', autoclose=True,
-                 handle_io_errors=None, mode=None):
+    def __init__(self,
+                 destination: TextIO | None = None,
+                 destination_path: StrPath | None = None,
+                 encoding: str | None = None,
+                 error_handler: str | None = 'strict',
+                 autoclose: bool = True,
+                 handle_io_errors: None = None,
+                 mode=None,
+                 ) -> None:
         """
         :Parameters:
             - `destination`: either a file-like object (which is written
@@ -446,8 +521,8 @@ class FileOutput(Output):
               `open`). The default is 'w', providing universal newline
               support for text files.
         """
-        Output.__init__(self, destination, destination_path,
-                        encoding, error_handler)
+        super().__init__(
+            destination, destination_path, encoding, error_handler)
         self.opened = True
         self.autoclose = autoclose
         if handle_io_errors is not None:
@@ -474,7 +549,7 @@ class FileOutput(Output):
             except AttributeError:
                 pass
 
-    def open(self):
+    def open(self) -> None:
         # Specify encoding
         if 'b' not in self.mode:
             kwargs = {'encoding': self.encoding,
@@ -488,20 +563,28 @@ class FileOutput(Output):
                               self.destination_path)
         self.opened = True
 
-    def write(self, data):
-        """Encode `data`, write it to a single file, and return it.
+    def write(self, data: str | bytes) -> str | bytes:
+        """Write `data` to a single file, also return it.
 
-        With Python 3 or binary output mode, `data` is returned unchanged,
-        except when specified encoding and output encoding differ.
+        `data` can be a `str` or `bytes` instance.
+        If writing `bytes` fails, an attempt is made to write to
+        the low-level interface ``self.destination.buffer``.
+
+        If `data` is a `str` instance and `self.encoding` and
+        `self.destination.encoding` are  set to different values, `data`
+        is encoded to a `bytes` instance using `self.encoding`.
+
+        Provisional: future versions may raise an error if `self.encoding`
+        and `self.destination.encoding` are set to different values.
         """
         if not self.opened:
             self.open()
-        if ('b' not in self.mode
+        if (isinstance(data, str)
             and check_encoding(self.destination, self.encoding) is False):
-            data = self.encode(data)
             if os.linesep != '\n':
-                # fix endings
-                data = data.replace(b'\n', bytes(os.linesep, 'ascii'))
+                data = data.replace('\n', os.linesep)  # fix endings
+            data = self.encode(data)
+
         try:
             self.destination.write(data)
         except TypeError as err:
@@ -512,23 +595,21 @@ class FileOutput(Output):
                     if check_encoding(self.destination,
                                       self.encoding) is False:
                         raise ValueError(
-                            'Encoding of %s (%s) differs \n'
-                            '  from specified encoding (%s)' %
-                            (self.destination_path or 'destination',
-                             self.destination.encoding,
-                             self.encoding))
+                            f'Encoding of {self.destination_path} '
+                            f'({self.destination.encoding}) differs \n'
+                            f'  from specified encoding ({self.encoding})')
                     else:
                         raise err
         except (UnicodeError, LookupError) as err:
             raise UnicodeError(
                 'Unable to encode output data. output-encoding is: '
-                '%s.\n(%s)' % (self.encoding, error_string(err)))
+                f'{self.encoding}.\n({error_string(err)})')
         finally:
             if self.autoclose:
                 self.close()
         return data
 
-    def close(self):
+    def close(self) -> None:
         if self.destination not in (sys.stdout, sys.stderr):
             self.destination.close()
             self.opened = False
@@ -537,63 +618,87 @@ class FileOutput(Output):
 class BinaryFileOutput(FileOutput):
     """
     A version of docutils.io.FileOutput which writes to a binary file.
+
+    Deprecated. Use `FileOutput` (works with `bytes` since Docutils 0.20).
+    Will be removed in Docutils 0.24.
     """
-    # Used by core.publish_cmdline_to_binary() which in turn is used by
-    # rst2odt (OpenOffice writer)
+    # Used by core.publish_cmdline_to_binary() which is also deprecated.
     mode = 'wb'
 
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        warnings.warn('"BinaryFileOutput" is obsoleted by "FileOutput"'
+                      ' and will be removed in Docutils 0.24.',
+                      DeprecationWarning, stacklevel=2)
+        super().__init__(*args, **kwargs)
+
 
 class StringInput(Input):
+    """Input from a `str` or `bytes` instance."""
 
-    """
-    Direct string input.
-    """
+    source: str | bytes
 
-    default_source_path = '<string>'
+    default_source_path: Final = '<string>'
 
-    def read(self):
-        """Decode and return the source string."""
+    def read(self) -> str:
+        """Return the source as `str` instance.
+
+        Decode, if required (see `Input.decode`).
+        """
         return self.decode(self.source)
 
 
 class StringOutput(Output):
+    """Output to a `bytes` or `str` instance.
 
-    """
-    Direct string output.
+    Provisional.
     """
 
-    default_destination_path = '<string>'
+    destination: str | bytes
 
-    def write(self, data):
-        """Encode `data`, store it in `self.destination`, and return it."""
+    default_destination_path: Final = '<string>'
+
+    def write(self, data: str | bytes) -> str | bytes:
+        """Store `data` in `self.destination`, and return it.
+
+        If `self.encoding` is set to the pseudo encoding name "unicode",
+        `data` must be a `str` instance and is stored/returned unchanged
+        (cf. `Output.encode`).
+
+        Otherwise, `data` can be a `bytes` or `str` instance and is
+        stored/returned as a `bytes` instance
+        (`str` data is encoded with `self.encode()`).
+
+        Attention: the `output_encoding`_ setting may affect the content
+        of the output (e.g. an encoding declaration in HTML or XML or the
+        representation of characters as LaTeX macro vs. literal character).
+        """
         self.destination = self.encode(data)
         return self.destination
 
 
 class NullInput(Input):
 
-    """
-    Degenerate input: read nothing.
-    """
+    """Degenerate input: read nothing."""
 
-    default_source_path = 'null input'
+    source: None
 
-    def read(self):
-        """Return a null string."""
+    default_source_path: Final = 'null input'
+
+    def read(self) -> str:
+        """Return an empty string."""
         return ''
 
 
 class NullOutput(Output):
 
-    """
-    Degenerate output: write nothing.
-    """
+    """Degenerate output: write nothing."""
 
-    default_destination_path = 'null output'
+    destination: None
 
-    def write(self, data):
-        """Do nothing ([don't even] send data to the bit bucket)."""
-        pass
+    default_destination_path: Final = 'null output'
+
+    def write(self, data: str | bytes) -> None:
+        """Do nothing, return None."""
 
 
 class DocTreeInput(Input):
@@ -604,8 +709,10 @@ class DocTreeInput(Input):
     The document tree must be passed in the ``source`` parameter.
     """
 
-    default_source_path = 'doctree input'
+    source: nodes.document
 
-    def read(self):
+    default_source_path: Final = 'doctree input'
+
+    def read(self) -> nodes.document:
         """Return the document tree."""
         return self.source
