@@ -1,21 +1,35 @@
-# $Id: __init__.py 8467 2020-01-26 21:23:42Z milde $
+# $Id: __init__.py 10046 2025-03-09 01:45:28Z aa-turner $
 # Author: David Goodger <goodger@python.org>
 # Copyright: This module has been placed in the public domain.
 
 # Internationalization details are documented in
-# <http://docutils.sf.net/docs/howto/i18n.html>.
+# <https://docutils.sourceforge.io/docs/howto/i18n.html>.
 
 """
 This package contains modules for language-dependent features of
 reStructuredText.
 """
 
+from __future__ import annotations
+
 __docformat__ = 'reStructuredText'
 
-import sys
-
 from docutils.languages import LanguageImporter
 
+TYPE_CHECKING = False
+if TYPE_CHECKING:
+    import types
+    from typing import NoReturn, Protocol, overload
+
+    class RSTLanguageModule(Protocol):
+        __name__: str
+
+        directives: dict[str, str]
+        roles: dict[str, str]
+else:
+    from docutils.utils._typing import overload
+
+
 class RstLanguageImporter(LanguageImporter):
     """Import language modules.
 
@@ -30,10 +44,20 @@ class RstLanguageImporter(LanguageImporter):
     warn_msg = 'rST localisation for language "%s" not found.'
     fallback = None
 
-    def check_content(self, module):
+    @overload
+    def check_content(self, module: RSTLanguageModule) -> None:
+        ...
+
+    @overload
+    def check_content(self, module: types.ModuleType) -> NoReturn:
+        ...
+
+    def check_content(self, module: RSTLanguageModule | types.ModuleType
+                      ) -> None:
         """Check if we got an rST language module."""
         if not (isinstance(module.directives, dict)
                 and isinstance(module.roles, dict)):
             raise ImportError
 
-get_language = RstLanguageImporter()
+
+get_language: LanguageImporter[RSTLanguageModule] = RstLanguageImporter()
