@@ -1,5 +1,4 @@
-# coding: utf-8
-# $Id: __init__.py 8295 2019-07-24 09:22:01Z grubert $
+# $Id: __init__.py 10250 2025-09-22 20:59:57Z milde $
 # Author: David Goodger <goodger@python.org>
 # Copyright: This module has been placed in the public domain.
 
@@ -7,30 +6,47 @@
 Miscellaneous utilities for the documentation utilities.
 """
 
+from __future__ import annotations
+
 __docformat__ = 'reStructuredText'
 
-import sys
+import itertools
 import os
 import os.path
 import re
-import itertools
-import warnings
+import sys
 import unicodedata
-from docutils import ApplicationError, DataError, __version_info__
-from docutils import nodes
-from docutils.nodes import unescape
-import docutils.io
-from docutils.utils.error_reporting import ErrorOutput, SafeString
+import warnings
+from pathlib import PurePath, Path
+
+from docutils import ApplicationError, DataError
+from docutils import io, nodes
+# for backwards compatibility
+from docutils.nodes import unescape  # noqa: F401 (imported but unused)
+
+TYPE_CHECKING = False
+if TYPE_CHECKING:
+    from collections.abc import Callable, Sequence, Iterable
+    from typing import Any, Final, Literal, TextIO
+
+    from docutils.utils._typing import TypeAlias
+
+    from docutils.nodes import StrPath
+    from docutils.frontend import Values
+
+    _ObserverFunc: TypeAlias = Callable[[nodes.system_message], None]
 
 
 class SystemMessage(ApplicationError):
 
-    def __init__(self, system_message, level):
+    def __init__(self, system_message: nodes.system_message, level: int,
+                 ) -> None:
         Exception.__init__(self, system_message.astext())
         self.level = level
 
 
-class SystemMessagePropagation(ApplicationError): pass
+class SystemMessagePropagation(ApplicationError):
+    pass
 
 
 class Reporter:
@@ -64,19 +80,32 @@ class Reporter:
        1995.
     """
 
-    levels = 'DEBUG INFO WARNING ERROR SEVERE'.split()
+    # Reporter.get_source_and_line is patched in by ``RSTState.runtime_init``
+    get_source_and_line: Callable[[int|None], tuple[StrPath|None, int|None]]
+
+    levels: Final[Sequence[str]] = (
+        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'SEVERE')
     """List of names for system message levels, indexed by level."""
 
     # system message level constants:
-    (DEBUG_LEVEL,
-     INFO_LEVEL,
-     WARNING_LEVEL,
-     ERROR_LEVEL,
-     SEVERE_LEVEL) = range(5)
+    DEBUG_LEVEL: Final = 0
+    INFO_LEVEL: Final = 1
+    WARNING_LEVEL: Final = 2
+    ERROR_LEVEL: Final = 3
+    SEVERE_LEVEL: Final = 4
+
+    def __init__(
+            self,
+            source: StrPath,
+            report_level: int,
+            halt_level: int,
+            stream: io.ErrorOutput|TextIO|str|Literal[False]|None = None,
+            debug: bool = False,
+            encoding: str|None = None,
+            error_handler: str = 'backslashreplace',
+            ) -> None:
+        """Low level instantiating. See also `new_reporter().`.
 
-    def __init__(self, source, report_level, halt_level, stream=None,
-                 debug=False, encoding=None, error_handler='backslashreplace'):
-        """
         :Parameters:
             - `source`: The path to or description of the source data.
             - `report_level`: The level at or above which warning output will
@@ -109,57 +138,50 @@ class Reporter:
         """The level at or above which `SystemMessage` exceptions
         will be raised, halting execution."""
 
-        if not isinstance(stream, ErrorOutput):
-            stream = ErrorOutput(stream, encoding, error_handler)
+        if not isinstance(stream, io.ErrorOutput):
+            stream = io.ErrorOutput(stream, encoding, error_handler)
 
-        self.stream = stream
+        self.stream: io.ErrorOutput = stream
         """Where warning output is sent."""
 
-        self.encoding = encoding or getattr(stream, 'encoding', 'ascii')
+        self.encoding: str = encoding or getattr(stream, 'encoding', 'ascii')
         """The output character encoding."""
 
-        self.observers = []
+        self.observers: list[_ObserverFunc] = []
         """List of bound methods or functions to call with each system_message
         created."""
 
-        self.max_level = -1
+        self.max_level: int = -1
         """The highest level system message generated so far."""
 
-    def set_conditions(self, category, report_level, halt_level,
-                       stream=None, debug=False):
-        warnings.warn('docutils.utils.Reporter.set_conditions deprecated; '
-                      'set attributes via configuration settings or directly',
-                      DeprecationWarning, stacklevel=2)
-        self.report_level = report_level
-        self.halt_level = halt_level
-        if not isinstance(stream, ErrorOutput):
-            stream = ErrorOutput(stream, self.encoding, self.error_handler)
-        self.stream = stream
-        self.debug_flag = debug
-
-    def attach_observer(self, observer):
+    def attach_observer(self, observer: _ObserverFunc) -> None:
         """
         The `observer` parameter is a function or bound method which takes one
         argument, a `nodes.system_message` instance.
         """
         self.observers.append(observer)
 
-    def detach_observer(self, observer):
+    def detach_observer(self, observer: _ObserverFunc) -> None:
         self.observers.remove(observer)
 
-    def notify_observers(self, message):
+    def notify_observers(self, message: nodes.system_message) -> None:
         for observer in self.observers:
             observer(message)
 
-    def system_message(self, level, message, *children, **kwargs):
+    def system_message(self,
+                       level: int,
+                       message: str,
+                       *children,
+                       **kwargs: Any
+                       ) -> nodes.system_message:
         """
         Return a system_message object.
 
         Raise an exception or generate a warning if appropriate.
         """
-        # `message` can be a `string`, `unicode`, or `Exception` instance.
+        # `message` can be a `str` or `Exception` instance.
         if isinstance(message, Exception):
-            message = SafeString(message)
+            message = str(message)
 
         attributes = kwargs.copy()
         if 'base_node' in kwargs:
@@ -169,11 +191,11 @@ class Reporter:
                 attributes.setdefault('source', source)
             if line is not None:
                 attributes.setdefault('line', line)
-                # assert source is not None, "node has line- but no source-argument"
-        if not 'source' in attributes: # 'line' is absolute line number
-            try: # look up (source, line-in-source)
+                # assert source is not None, "line- but no source-argument"
+        if 'source' not in attributes:
+            # 'line' is absolute line number
+            try:
                 source, line = self.get_source_and_line(attributes.get('line'))
-                # print "locator lookup", kwargs.get('line'), "->", source, line
             except AttributeError:
                 source, line = None, None
             if source is not None:
@@ -198,52 +220,60 @@ class Reporter:
         self.max_level = max(level, self.max_level)
         return msg
 
-    def debug(self, *args, **kwargs):
+    def debug(self, *args, **kwargs: Any) -> nodes.system_message:
         """
-        Level-0, "DEBUG": an internal reporting issue. Typically, there is no
-        effect on the processing. Level-0 system messages are handled
-        separately from the others.
+        Level-0, "DEBUG": an internal reporting issue.
+
+        Typically, there is no effect on the processing. Level-0 system
+        messages are handled separately from the others.
         """
         if self.debug_flag:
             return self.system_message(self.DEBUG_LEVEL, *args, **kwargs)
 
-    def info(self, *args, **kwargs):
+    def info(self, *args, **kwargs: Any) -> nodes.system_message:
         """
-        Level-1, "INFO": a minor issue that can be ignored. Typically there is
-        no effect on processing, and level-1 system messages are not reported.
+        Level-1, "INFO": a minor issue that can be ignored.
+
+        Typically, there is no effect on processing and level-1 system
+        messages are not reported.
         """
         return self.system_message(self.INFO_LEVEL, *args, **kwargs)
 
-    def warning(self, *args, **kwargs):
+    def warning(self, *args, **kwargs: Any) -> nodes.system_message:
         """
-        Level-2, "WARNING": an issue that should be addressed. If ignored,
-        there may be unpredictable problems with the output.
+        Level-2, "WARNING": an issue that should be addressed.
+
+        If ignored, there may be unpredictable problems with the output.
         """
         return self.system_message(self.WARNING_LEVEL, *args, **kwargs)
 
-    def error(self, *args, **kwargs):
+    def error(self, *args, **kwargs: Any) -> nodes.system_message:
         """
-        Level-3, "ERROR": an error that should be addressed. If ignored, the
-        output will contain errors.
+        Level-3, "ERROR": an error that should be addressed.
+
+        If ignored, the output will contain errors.
         """
         return self.system_message(self.ERROR_LEVEL, *args, **kwargs)
 
-    def severe(self, *args, **kwargs):
+    def severe(self, *args, **kwargs: Any) -> nodes.system_message:
         """
-        Level-4, "SEVERE": a severe error that must be addressed. If ignored,
-        the output will contain severe errors. Typically level-4 system
-        messages are turned into exceptions which halt processing.
+        Level-4, "SEVERE": a severe error that must be addressed.
+
+        If ignored, the output will contain severe errors. Typically level-4
+        system messages are turned into exceptions which halt processing.
         """
         return self.system_message(self.SEVERE_LEVEL, *args, **kwargs)
 
 
-class ExtensionOptionError(DataError): pass
-class BadOptionError(ExtensionOptionError): pass
-class BadOptionDataError(ExtensionOptionError): pass
-class DuplicateOptionError(ExtensionOptionError): pass
+class ExtensionOptionError(DataError): pass  # NoQA: E701
+class BadOptionError(ExtensionOptionError): pass  # NoQA: E701
+class BadOptionDataError(ExtensionOptionError): pass  # NoQA: E701
+class DuplicateOptionError(ExtensionOptionError): pass  # NoQA: E701
 
 
-def extract_extension_options(field_list, options_spec):
+def extract_extension_options(field_list: nodes.field_list,
+                              options_spec: dict[str, Callable[object], Any],
+                              ) -> dict[str, Any]:
     """
     Return a dictionary mapping extension option names to converted values.
 
@@ -265,10 +295,11 @@ def extract_extension_options(field_list, options_spec):
           missing data, bad quotes, etc.).
     """
     option_list = extract_options(field_list)
-    option_dict = assemble_option_dict(option_list, options_spec)
-    return option_dict
+    return assemble_option_dict(option_list, options_spec)
 
-def extract_options(field_list):
+
+def extract_options(field_list: nodes.field_list
+                    ) -> list[tuple[str, str|None]]:
     """
     Return a list of option (name, value) pairs from field names & bodies.
 
@@ -290,8 +321,10 @@ def extract_options(field_list):
         body = field[1]
         if len(body) == 0:
             data = None
-        elif len(body) > 1 or not isinstance(body[0], nodes.paragraph) \
-              or len(body[0]) != 1 or not isinstance(body[0][0], nodes.Text):
+        elif (len(body) > 1
+              or not isinstance(body[0], nodes.paragraph)
+              or len(body[0]) != 1
+              or not isinstance(body[0][0], nodes.Text)):
             raise BadOptionDataError(
                   'extension option field body may contain\n'
                   'a single paragraph only (option "%s")' % name)
@@ -300,7 +333,10 @@ def extract_options(field_list):
         option_list.append((name, data))
     return option_list
 
-def assemble_option_dict(option_list, options_spec):
+
+def assemble_option_dict(option_list: list[tuple[str, str|None]],
+                         options_spec: dict[str, Callable[object], Any],
+                         ) -> dict[str, Any]:
     """
     Return a mapping of option names to values.
 
@@ -327,7 +363,7 @@ def assemble_option_dict(option_list, options_spec):
             raise DuplicateOptionError('duplicate option "%s"' % name)
         try:
             options[name] = convertor(value)
-        except (ValueError, TypeError), detail:
+        except (ValueError, TypeError) as detail:
             raise detail.__class__('(option: "%s"; value: %r)\n%s'
                                    % (name, value, ' '.join(detail.args)))
     return options
@@ -336,25 +372,29 @@ def assemble_option_dict(option_list, options_spec):
 class NameValueError(DataError): pass
 
 
-def decode_path(path):
+def decode_path(path: str|bytes|None) -> str:
     """
-    Ensure `path` is Unicode. Return `nodes.reprunicode` object.
+    Ensure `path` is Unicode. Return `str` instance.
 
-    Decode file/path string in a failsave manner if not already done.
+    Decode file/path string in a failsafe manner if not already done.
+
+    Deprecated. Will be removed in Docutils 1.0.
     """
-    # see also http://article.gmane.org/gmane.text.docutils.user/2905
-    if isinstance(path, unicode):
+    if isinstance(path, str):
         return path
+    if path is None:
+        return ''
     try:
         path = path.decode(sys.getfilesystemencoding(), 'strict')
-    except AttributeError: # default value None has no decode method
-        return nodes.reprunicode(path)
+    except AttributeError:
+        raise ValueError('`path` value must be a String or ``None``, '
+                         f'not {path!r}')
     except UnicodeDecodeError:
         try:
             path = path.decode('utf-8', 'strict')
         except UnicodeDecodeError:
             path = path.decode('ascii', 'replace')
-    return nodes.reprunicode(path)
+    return path
 
 
 def extract_name_value(line):
@@ -367,41 +407,39 @@ def extract_name_value(line):
     """
     attlist = []
     while line:
-        equals = line.find('=')
-        if equals == -1:
+        equals_index = line.find('=')
+        if equals_index == -1:
             raise NameValueError('missing "="')
-        attname = line[:equals].strip()
-        if equals == 0 or not attname:
-            raise NameValueError(
-                  'missing attribute name before "="')
-        line = line[equals+1:].lstrip()
+        attname = line[:equals_index].strip()
+        if equals_index == 0 or not attname:
+            raise NameValueError('missing attribute name before "="')
+        line = line[equals_index+1:].lstrip()
         if not line:
-            raise NameValueError(
-                  'missing value after "%s="' % attname)
+            raise NameValueError(f'missing value after "{attname}="')
         if line[0] in '\'"':
-            endquote = line.find(line[0], 1)
-            if endquote == -1:
+            endquote_index = line.find(line[0], 1)
+            if endquote_index == -1:
                 raise NameValueError(
-                      'attribute "%s" missing end quote (%s)'
-                      % (attname, line[0]))
-            if len(line) > endquote + 1 and line[endquote + 1].strip():
-                raise NameValueError(
-                      'attribute "%s" end quote (%s) not followed by '
-                      'whitespace' % (attname, line[0]))
-            data = line[1:endquote]
-            line = line[endquote+1:].lstrip()
+                      f'attribute "{attname}" missing end quote ({line[0]})')
+            if (len(line) > endquote_index + 1
+                and line[endquote_index + 1].strip()):
+                raise NameValueError(f'attribute "{attname}" end quote '
+                                     f'({line[0]}) not followed by whitespace')
+            data = line[1:endquote_index]
+            line = line[endquote_index+1:].lstrip()
         else:
-            space = line.find(' ')
-            if space == -1:
+            space_index = line.find(' ')
+            if space_index == -1:
                 data = line
                 line = ''
             else:
-                data = line[:space]
-                line = line[space+1:].lstrip()
+                data = line[:space_index]
+                line = line[space_index+1:].lstrip()
         attlist.append((attname.lower(), data))
     return attlist
 
-def new_reporter(source_path, settings):
+
+def new_reporter(source_path: StrPath, settings: Values) -> Reporter:
     """
     Return a new Reporter object.
 
@@ -418,34 +456,48 @@ def new_reporter(source_path, settings):
         error_handler=settings.error_encoding_error_handler)
     return reporter
 
-def new_document(source_path, settings=None):
+
+def new_document(source_path: StrPath, settings: Values|None = None
+                 ) -> nodes.document:
     """
     Return a new empty document object.
 
     :Parameters:
-        `source_path` : string
+        `source_path` : str or pathlib.Path
             The path to or description of the source text of the document.
         `settings` : optparse.Values object
             Runtime settings.  If none are provided, a default core set will
             be used.  If you will use the document object with any Docutils
-            components, you must provide their default settings as well.  For
-            example, if parsing, at least provide the parser settings,
-            obtainable as follows::
+            components, you must provide their default settings as well.
+
+            For example, if parsing rST, at least provide the rst-parser
+            settings, obtainable as follows:
+
+            Defaults for parser component::
+
+                settings = docutils.frontend.get_default_settings(
+                               docutils.parsers.rst.Parser)
+
+            Defaults and configuration file customizations::
+
+                settings = docutils.core.Publisher(
+                    parser=docutils.parsers.rst.Parser).get_settings()
 
-                settings = docutils.frontend.OptionParser(
-                    components=(docutils.parsers.rst.Parser,)
-                    ).get_default_values()
     """
+    # Import at top of module would lead to circular dependency!
     from docutils import frontend
     if settings is None:
-        settings = frontend.OptionParser().get_default_values()
-    source_path = decode_path(source_path)
+        settings = frontend.get_default_settings()
     reporter = new_reporter(source_path, settings)
     document = nodes.document(settings, reporter, source=source_path)
     document.note_source(source_path, -1)
     return document
 
-def clean_rcs_keywords(paragraph, keyword_substitutions):
+
+def clean_rcs_keywords(
+    paragraph: nodes.paragraph,
+    keyword_substitutions: Sequence[tuple[re.Pattern[[str], str]]],
+) -> None:
     if len(paragraph) == 1 and isinstance(paragraph[0], nodes.Text):
         textnode = paragraph[0]
         for pattern, substitution in keyword_substitutions:
@@ -454,14 +506,44 @@ def clean_rcs_keywords(paragraph, keyword_substitutions):
                 paragraph[0] = nodes.Text(pattern.sub(substitution, textnode))
                 return
 
-def relative_path(source, target):
+
+def relative_path(source: StrPath|None, target: StrPath) -> str:
     """
     Build and return a path to `target`, relative to `source` (both files).
 
-    If there is no common prefix, return the absolute path to `target`.
+    The return value is a `str` suitable to be included in `source`
+    as a reference to `target`.
+
+    :Parameters:
+        `source` : path-like object or None
+            Path of a file in the start directory for the relative path
+            (the file does not need to exist).
+            The value ``None`` is replaced with "<cwd>/dummy_file".
+        `target` : path-like object
+            End point of the returned relative path.
+
+    Differences to `os.path.relpath()`:
+
+    * Inverse argument order.
+    * `source` is assumed to be a FILE in the start directory (add a "dummy"
+      file name to obtain the path relative from a directory)
+      while `os.path.relpath()` expects a DIRECTORY as `start` argument.
+    * Always use Posix path separator ("/") for the output.
+    * Use `os.sep` for parsing the input
+      (changing the value of `os.sep` is ignored by `os.relpath()`).
+    * If there is no common prefix, return the absolute path to `target`.
+
+    Differences to `pathlib.PurePath.relative_to(other)`:
+
+    * pathlib offers an object oriented interface.
+    * `source` expects path to a FILE while `other` expects a DIRECTORY.
+    * `target` defaults to the cwd, no default value for `other`.
+    * `relative_path()` always returns a path (relative or absolute),
+      while `PurePath.relative_to()` raises a ValueError
+      if `target` is not a subpath of `other` (no ".." inserted).
     """
     source_parts = os.path.abspath(source or type(target)('dummy_file')
-                                  ).split(os.sep)
+                                   ).split(os.sep)
     target_parts = os.path.abspath(target).split(os.sep)
     # Check first 2 parts because '/dir'.split('/') == ['', 'dir']:
     if source_parts[:2] != target_parts[:2]:
@@ -479,66 +561,73 @@ def relative_path(source, target):
     parts = ['..'] * (len(source_parts) - 1) + target_parts
     return '/'.join(parts)
 
-def get_stylesheet_reference(settings, relative_to=None):
+
+def get_stylesheet_reference(settings: Values,
+                             relative_to: StrPath|None = None
+                             ) -> str:
     """
     Retrieve a stylesheet reference from the settings object.
 
-    Deprecated. Use get_stylesheet_list() instead to
-    enable specification of multiple stylesheets as a comma-separated
-    list.
+    Deprecated. Will be removed in Docutils 1.0.
+    Use get_stylesheet_list() instead to enable specification of multiple
+    stylesheets as a comma-separated list.
     """
+    warnings.warn('utils.get_stylesheet_reference()'
+                  ' is obsoleted by utils.get_stylesheet_list()'
+                  ' and will be removed in Docutils 2.0.',
+                  DeprecationWarning, stacklevel=2)
     if settings.stylesheet_path:
         assert not settings.stylesheet, (
             'stylesheet and stylesheet_path are mutually exclusive.')
-        if relative_to == None:
-            relative_to = settings._destination
+        if relative_to is None:
+            relative_to = settings.output_path
         return relative_path(relative_to, settings.stylesheet_path)
     else:
         return settings.stylesheet
 
+
 # Return 'stylesheet' or 'stylesheet_path' arguments as list.
 #
 # The original settings arguments are kept unchanged: you can test
-# with e.g. ``if settings.stylesheet_path:``
+# with e.g. ``if settings.stylesheet_path: ...``.
 #
-# Differences to ``get_stylesheet_reference``:
+# Differences to the depracated `get_stylesheet_reference()`:
 # * return value is a list
 # * no re-writing of the path (and therefore no optional argument)
 #   (if required, use ``utils.relative_path(source, target)``
 #   in the calling script)
-def get_stylesheet_list(settings):
-    """
-    Retrieve list of stylesheet references from the settings object.
-    """
+def get_stylesheet_list(settings: Values) -> list[str]:
+    """Retrieve list of stylesheet references from the settings object."""
     assert not (settings.stylesheet and settings.stylesheet_path), (
             'stylesheet and stylesheet_path are mutually exclusive.')
     stylesheets = settings.stylesheet_path or settings.stylesheet or []
-    # programmatically set default can be string or unicode:
+    # programmatically set default may be string with comma separated list:
     if not isinstance(stylesheets, list):
         stylesheets = [path.strip() for path in stylesheets.split(',')]
-    # expand relative paths if found in stylesheet-dirs:
-    return [find_file_in_dirs(path, settings.stylesheet_dirs)
-            for path in stylesheets]
+    if settings.stylesheet_path:
+        # expand relative paths if found in stylesheet-dirs:
+        stylesheets = [find_file_in_dirs(path, settings.stylesheet_dirs)
+                       for path in stylesheets]
+    return stylesheets
 
-def find_file_in_dirs(path, dirs):
+
+def find_file_in_dirs(path: StrPath, dirs: Iterable[StrPath]) -> str:
     """
     Search for `path` in the list of directories `dirs`.
 
     Return the first expansion that matches an existing file.
     """
-    if os.path.isabs(path):
-        return path
+    path = Path(path)
+    if path.is_absolute():
+        return path.as_posix()
     for d in dirs:
-        if d == '.':
-            f = path
-        else:
-            d = os.path.expanduser(d)
-            f = os.path.join(d, path)
-        if os.path.exists(f):
-            return f
-    return path
+        f = Path(d).expanduser() / path
+        if f.exists():
+            return f.as_posix()
+    return path.as_posix()
 
-def get_trim_footnote_ref_space(settings):
+
+def get_trim_footnote_ref_space(settings: Values) -> bool:
     """
     Return whether or not to trim footnote space.
 
@@ -547,13 +636,13 @@ def get_trim_footnote_ref_space(settings):
     If trim_footnote_reference_space is None, return False unless the
     footnote reference style is 'superscript'.
     """
-    if settings.trim_footnote_reference_space is None:
-        return hasattr(settings, 'footnote_references') and \
-               settings.footnote_references == 'superscript'
+    if settings.setdefault('trim_footnote_reference_space', None) is None:
+        return getattr(settings, 'footnote_references', None) == 'superscript'
     else:
         return settings.trim_footnote_reference_space
 
-def get_source_line(node):
+
+def get_source_line(node) -> tuple[StrPath|None, int|None]:
     """
     Return the "source" and "line" attributes from the `node` given or from
     its closest ancestor.
@@ -564,22 +653,22 @@ def get_source_line(node):
         node = node.parent
     return None, None
 
-def escape2null(text):
+
+def escape2null(text: str) -> str:
     """Return a string with escape-backslashes converted to nulls."""
     parts = []
     start = 0
     while True:
-        found = text.find('\\', start)
-        if found == -1:
+        bs_index = text.find('\\', start)
+        if bs_index == -1:
             parts.append(text[start:])
             return ''.join(parts)
-        parts.append(text[start:found])
-        parts.append('\x00' + text[found+1:found+2])
-        start = found + 2               # skip character after escape
+        parts.extend((text[start:bs_index],
+                      '\x00' + text[bs_index + 1:bs_index + 2]))
+        start = bs_index + 2  # skip character after escape
 
-# `unescape` definition moved to `nodes` to avoid circular import dependency.
 
-def split_escaped_whitespace(text):
+def split_escaped_whitespace(text: str) -> list[str]:
     """
     Split `text` on escaped whitespace (null+space or null+newline).
     Return a list of strings.
@@ -589,76 +678,67 @@ def split_escaped_whitespace(text):
     # flatten list of lists of strings to list of strings:
     return list(itertools.chain(*strings))
 
-def strip_combining_chars(text):
-    if isinstance(text, str) and sys.version_info < (3,0):
-        return text
-    return u''.join([c for c in text if not unicodedata.combining(c)])
 
-def find_combining_chars(text):
+def strip_combining_chars(text: str) -> str:
+    return ''.join(c for c in text if not unicodedata.combining(c))
+
+
+def find_combining_chars(text: str) -> list[int]:
     """Return indices of all combining chars in  Unicode string `text`.
 
     >>> from docutils.utils import find_combining_chars
-    >>> find_combining_chars(u'A t̆ab̆lĕ')
+    >>> find_combining_chars('A t̆ab̆lĕ')
     [3, 6, 9]
 
     """
-    if isinstance(text, str) and sys.version_info < (3,0):
-        return []
-    return [i for i,c in enumerate(text) if unicodedata.combining(c)]
+    return [i for i, c in enumerate(text) if unicodedata.combining(c)]
 
-def column_indices(text):
+
+def column_indices(text: str) -> list[int]:
     """Indices of Unicode string `text` when skipping combining characters.
 
     >>> from docutils.utils import column_indices
-    >>> column_indices(u'A t̆ab̆lĕ')
+    >>> column_indices('A t̆ab̆lĕ')
     [0, 1, 2, 4, 5, 7, 8]
 
     """
     # TODO: account for asian wide chars here instead of using dummy
     # replacements in the tableparser?
-    string_indices = range(len(text))
-    for index in find_combining_chars(text):
-        string_indices[index] = None
-    return [i for i in string_indices if i is not None]
+    return [i for i, c in enumerate(text) if not unicodedata.combining(c)]
+
 
 east_asian_widths = {'W': 2,   # Wide
                      'F': 2,   # Full-width (wide)
                      'Na': 1,  # Narrow
                      'H': 1,   # Half-width (narrow)
                      'N': 1,   # Neutral (not East Asian, treated as narrow)
-                     'A': 1}   # Ambiguous (s/b wide in East Asian context,
-                               # narrow otherwise, but that doesn't work)
+                     'A': 1,   # Ambiguous (s/b wide in East Asian context,
+                     }         # narrow otherwise, but that doesn't work)
 """Mapping of result codes from `unicodedata.east_asian_widt()` to character
 column widths."""
 
-def column_width(text):
+
+def column_width(text: str) -> int:
     """Return the column width of text.
 
     Correct ``len(text)`` for wide East Asian and combining Unicode chars.
     """
-    if isinstance(text, str) and sys.version_info < (3,0):
-        return len(text)
-    width = sum([east_asian_widths[unicodedata.east_asian_width(c)]
-                 for c in text])
+    width = sum(east_asian_widths[unicodedata.east_asian_width(c)]
+                for c in text)
     # correction for combining chars:
     width -= len(find_combining_chars(text))
     return width
 
-def uniq(L):
-     r = []
-     for item in L:
-         if not item in r:
-             r.append(item)
-     return r
 
-def unique_combinations(items, n):
-    """Return `itertools.combinations`."""
-    warnings.warn('docutils.utils.unique_combinations is deprecated; '
-                  'use itertools.combinations directly.',
-                      DeprecationWarning, stacklevel=2)
-    return itertools.combinations(items, n)
+def uniq(L: list) -> list:
+    r = []
+    for item in L:
+        if item not in r:
+            r.append(item)
+    return r
 
-def normalize_language_tag(tag):
+
+def normalize_language_tag(tag: str) -> list[str]:
     """Return a list of normalized combinations for a `BCP 47` language tag.
 
     Example:
@@ -671,22 +751,34 @@ def normalize_language_tag(tag):
 
     """
     # normalize:
-    tag = tag.lower().replace('-','_')
+    tag = tag.lower().replace('-', '_')
     # split (except singletons, which mark the following tag as non-standard):
     tag = re.sub(r'_([a-zA-Z0-9])_', r'_\1-', tag)
-    subtags = [subtag for subtag in tag.split('_')]
+    subtags = list(tag.split('_'))
     base_tag = (subtags.pop(0),)
     # find all combinations of subtags
-    taglist = []
-    for n in range(len(subtags), 0, -1):
-        # for tags in unique_combinations(subtags, n):
-        for tags in itertools.combinations(subtags, n):
-            taglist.append('-'.join(base_tag+tags))
+    taglist = ['-'.join(base_tag + tags)
+               for n in range(len(subtags), 0, -1)
+               for tags in itertools.combinations(subtags, n)
+               ]
     taglist += base_tag
     return taglist
 
 
-class DependencyList(object):
+def xml_declaration(encoding: str|Literal['unicode']|None = None) -> str:
+    """Return an XML text declaration.
+
+    Include an encoding declaration, if `encoding`
+    is not 'unicode', '', or None.
+    """
+    if encoding and encoding.lower() != 'unicode':
+        encoding_declaration = f' encoding="{encoding}"'
+    else:
+        encoding_declaration = ''
+    return f'<?xml version="1.0"{encoding_declaration}?>\n'
+
+
+class DependencyList:
 
     """
     List of dependencies, with file recording support.
@@ -695,101 +787,64 @@ class DependencyList(object):
     to explicitly call the close() method.
     """
 
-    def __init__(self, output_file=None, dependencies=[]):
+    def __init__(self,
+                 output_file: Literal['-'] | StrPath | None = None,
+                 dependencies: Iterable[StrPath] = ()
+                 ) -> None:
         """
         Initialize the dependency list, automatically setting the
         output file to `output_file` (see `set_output()`) and adding
         all supplied dependencies.
+
+        If output_file is None, no file output is done when calling add().
         """
         self.set_output(output_file)
-        for i in dependencies:
-            self.add(i)
+        self.add(*dependencies)
 
-    def set_output(self, output_file):
+    def set_output(self, output_file: Literal['-']|StrPath|None) -> None:
         """
         Set the output file and clear the list of already added
         dependencies.
 
-        `output_file` must be a string.  The specified file is
-        immediately overwritten.
+        The specified file is immediately overwritten.
 
-        If output_file is '-', the output will be written to stdout.
-        If it is None, no file output is done when calling add().
+        If `output_file` is '-', the output will be written to stdout.
+        The empty string or None stop output.
         """
-        self.list = []
-        if output_file:
-            if output_file == '-':
-                of = None
-            else:
-                of = output_file
-            self.file = docutils.io.FileOutput(destination_path=of,
-                                   encoding='utf8', autoclose=False)
+        if output_file == '-':
+            self.file = sys.stdout
+        elif output_file:
+            self.file = open(output_file, 'w', encoding='utf-8')
         else:
             self.file = None
+        self.list = []
 
-    def add(self, *filenames):
+    def add(self, *paths: StrPath) -> None:
         """
-        If the dependency `filename` has not already been added,
-        append it to self.list and print it to self.file if self.file
-        is not None.
+        Append `path` to `self.list` unless it is already there.
+
+        Also append to `self.file` unless it is already there
+        or `self.file is `None`.
         """
-        for filename in filenames:
-            if not filename in self.list:
-                self.list.append(filename)
+        for path in paths:
+            if isinstance(path, PurePath):
+                path = path.as_posix()  # use '/' as separator
+            if path not in self.list:
+                self.list.append(path)
                 if self.file is not None:
-                    self.file.write(filename+'\n')
+                    self.file.write(path+'\n')
 
-    def close(self):
+    def close(self) -> None:
         """
         Close the output file.
         """
-        self.file.close()
+        if self.file is not sys.stdout:
+            self.file.close()
         self.file = None
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         try:
             output_file = self.file.name
         except AttributeError:
             output_file = None
         return '%s(%r, %s)' % (self.__class__.__name__, output_file, self.list)
-
-
-release_level_abbreviations = {
-    'alpha':     'a',
-    'beta':      'b',
-    'candidate': 'rc',
-    'final':     '',}
-
-def version_identifier(version_info=None):
-    """
-    Return a version identifier string built from `version_info`, a
-    `docutils.VersionInfo` namedtuple instance or compatible tuple. If
-    `version_info` is not provided, by default return a version identifier
-    string based on `docutils.__version_info__` (i.e. the current Docutils
-    version).
-    """
-    if version_info is None:
-        version_info = __version_info__
-    if version_info.micro:
-        micro = '.%s' % version_info.micro
-    else:
-        # 0 is omitted:
-        micro = ''
-    releaselevel = release_level_abbreviations[version_info.releaselevel]
-    if version_info.serial:
-        serial = version_info.serial
-    else:
-        # 0 is omitted:
-        serial = ''
-    if version_info.release:
-        dev = ''
-    else:
-        dev = '.dev'
-    version = '%s.%s%s%s%s%s' % (
-        version_info.major,
-        version_info.minor,
-        micro,
-        releaselevel,
-        serial,
-        dev)
-    return version
