@@ -1,9 +1,9 @@
 # Source and destination file names
-test_source = "standalone_rst_html5.txt"
+test_source = "standalone_rst_html5.rst"
 test_destination = "standalone_rst_html5.html"
 
 # Keyword parameters passed to publish_file()
-writer_name = "html5"
+writer = "html5"
 settings_overrides = {
     'sectsubtitle_xform': True,
     # "smart" quotes:
