@@ -13,8 +13,15 @@
 Various tests for the `pygments` code highlighter.
 """
 
+from pathlib import Path
+import sys
 import unittest
 
+if __name__ == '__main__':
+    # prepend the "docutils root" to the Python library path
+    # so we import the local `docutils` package.
+    sys.path.insert(0, str(Path(__file__).resolve().parents[4]))
+
 from docutils.core import publish_string
 from docutils.utils.code_analyzer import with_pygments
 
