@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# $Id: test_null.py 9425 2023-06-30 14:56:47Z milde $
+# $Id: test_null.py 9914 2024-08-21 13:45:08Z milde $
 # Author: Lea Wiemann <LeWiemann@gmail.com>
 # Copyright: This module has been placed in the public domain.
 
@@ -18,17 +18,17 @@ if __name__ == '__main__':
     sys.path.insert(0, str(Path(__file__).resolve().parents[2]))
 
 from docutils.core import publish_string
+from docutils.writers import null
 
 
 class WriterPublishTestCase(unittest.TestCase):
     def test_publish(self):
-        writer_name = 'null'
         for name, cases in totest.items():
             for casenum, (case_input, case_expected) in enumerate(cases):
                 with self.subTest(id=f'totest[{name!r}][{casenum}]'):
                     output = publish_string(
                         source=case_input,
-                        writer_name=writer_name,
+                        writer=null.Writer(),
                         settings_overrides={
                             '_disable_config': True,
                             'strict_visitor': True,
@@ -45,7 +45,7 @@ totest['basic'] = [
 ["""\
 This is a paragraph.
 """,
-None]
+'']
 ]
 
 if __name__ == '__main__':
